"""
Methods for groups
"""
from typing import Union
from pyfreeipa.Api._utils import delist, listdelist


def group(
        self,
        cn: str
):
    """
    @brief      Returns only the group response

    @param      self  The object
    @param      cn   The cn of the group to return

    @return     the group account as a dictionary
    """

    group = None
    response = self.group_show(cn, allattrs=True)

    if response.json()['result']:
        group = delist(response.json()['result']['result'])

    return group


def groups(
        self,
        searchstring: Union[str, None]=None,
        cn: Union[str, None]=None,
        gidnumber: Union[int, None]=None,
        in_group: Union[str, list, None]=None
):
    """
    @brief      Given a some search parameters find all the group acounts that match

    @param      self          The object
    @param      searchstring  The searchstring used on any otptoken attribute
    @param      uniqueid      substring used to match otptoken uniqueid
    @param      owner         search for tokens owned but specified group

    @return     { description_of_the_return_value }
    """

    grouplist = []
    response = self.group_find(
        searchstring=searchstring,
        cn=cn,
        gidnumber=gidnumber,
        in_group=in_group,
        allattrs=True
    )

    if response.json()['result']:
        grouplist = listdelist(response.json()['result']['result'])

    return grouplist


def grouplist(
        self,
        groups: Union[str, list, None]=None
):
    """
    @brief      Given a list of gids and/or groups, return a list of groupnames that match or are members

    @param      self          The object
    @param      searchstring  The searchstring used on any otptoken attribute
    @param      uniqueid      substring used to match otptoken uniqueid
    @param      owner         search for tokens owned but specified group

    @return     { description_of_the_return_value }
    """

    grouplist = []

    if groups:
        if isinstance(groups, list):
            for groupname in groups:
                response = self.group(groupname)
                if response:
                    grouplist.append(groupname)
                    response = self.groups(in_group=groupname)
                    if isinstance(response, list):
                        for group in response:
                            grouplist.append(group['cn'])
                    else:
                        grouplist.append(response['cn'])
        else:
            response = self.group(groupname)
            if response:
                grouplist.append(groupname)
                response = self.groups(in_group=groups)
                if isinstance(response, list):
                    for group in response:
                        grouplist.append(group['cn'])
                else:
                    grouplist.append(response['cn'])
    else:
        for group in self.groups():
            grouplist.append(group['cn'])

    if len(grouplist) > 1:
        grouplist = sorted(set(grouplist))

    return grouplist


def group_show(
        self,
        cn: str,
        no_members: Union[bool, None]=None,
        rights: Union[bool, None]=None,
        allattrs: Union[bool, None]=None,
        raw: Union[bool, None]=None
):
    """
    @brief      A complete implementation of the group_show command

    @param      self  The object

    @param      cn of the group to be shown

    @param      rights, if true, displays the access rights of this group

    @param      all, retrieves all attributes

    @param      raw, returns the raw response, only changes output format

    @return     the requests.Response from the ping request
    """

    method = 'group_show'

    args = cn

    params = {}

    if rights is not None:
        params['rights'] = rights

    if allattrs is not None:
        params['all'] = allattrs

    if raw is not None:
        params['raw'] = raw

    if no_members is not None:
        params['no_members'] = no_members

    return self.request(
        method,
        args=args,
        params=params
    )


def group_find(
        self,
        searchstring: Union[str, None]=None,
        cn: Union[str, None]=None,
        gidnumber: Union[int, None]=None,
        private: Union[bool, None]=None,
        posix: Union[bool, None]=None,
        external: Union[bool, None]=None,
        nonposix: Union[bool, None]=None,
        user: Union[str, list, None]=None,
        no_user: Union[str, list, None]=None,
        group: Union[str, list, None]=None,
        no_group: Union[str, list, None]=None,
        in_group: Union[str, list, None]=None,
        not_in_group: Union[str, list, None]=None,
        in_netgroup: Union[str, list, None]=None,
        not_in_netgroup: Union[str, list, None]=None,
        in_role: Union[str, list, None]=None,
        not_in_role: Union[str, list, None]=None,
        in_hbacrule: Union[str, list, None]=None,
        not_in_hbacrule: Union[str, list, None]=None,
        in_sudorule: Union[str, list, None]=None,
        not_in_sudorule: Union[str, list, None]=None,
        no_members: Union[bool, None]=None,
        rights: Union[bool, None]=None,
        allattrs: Union[bool, None]=None,
        raw: Union[bool, None]=None,
        pkey_only: Union[bool, None]=None
):
    """
    @brief      A partial implementation of the group_find request

    @param      self  The object

    @param      gid of the group to be shown

    @param      rights, if true, displays the access rights of this group

    @param      all, retrieves all attributes

    @param      raw, returns the raw response, only changes output format

    @return     the requests.Response from the ping request
    """

    method = 'group_find'

    args = None

    if searchstring:
        args = searchstring

    params = {}

    if cn is not None:
        params['cn'] = cn

    if gidnumber is not None:
        params['gidnumber'] = gidnumber

    if private is not None:
        params['private'] = private

    if posix is not None:
        params['posix'] = posix

    if external is not None:
        params['external'] = external

    if nonposix is not None:
        params['nonposix'] = nonposix

    if user is not None:
        params['user'] = user

    if no_user is not None:
        params['no_user'] = no_user

    if group is not None:
        params['group'] = group

    if no_group is not None:
        params['no_group'] = no_group

    if in_group is not None:
        params['in_group'] = in_group

    if not_in_group is not None:
        params['not_in_group'] = not_in_group

    if in_netgroup is not None:
        params['in_netgroup'] = in_netgroup

    if not_in_netgroup is not None:
        params['not_in_netgroup'] = not_in_netgroup

    if in_role is not None:
        params['in_role'] = in_role

    if not_in_role is not None:
        params['not_in_role'] = not_in_role

    if in_hbacrule is not None:
        params['in_hbacrule'] = in_hbacrule

    if not_in_hbacrule is not None:
        params['not_in_hbacrule'] = not_in_hbacrule

    if in_sudorule is not None:
        params['in_sudorule'] = in_sudorule

    if not_in_sudorule is not None:
        params['not_in_sudorule'] = not_in_sudorule

    if no_members is not None:
        params['no_members'] = no_members

    if rights is not None:
        params['rights'] = rights

    if allattrs is not None:
        params['all'] = allattrs

    if raw is not None:
        params['raw'] = raw

    if pkey_only is not None:
        params['pkey_only'] = pkey_only

    return self.request(
        method,
        args=args,
        params=params
    )



"""
Methods for users
"""
from datetime import datetime
from typing import Union
from pyfreeipa.Api._utils import delist, listdelist

def user(
        self,
        uid: str
):
    """
    @brief      Returns only the user response

    @param      self  The object
    @param      uid   The uid of the user to return

    @return     the user account as a dictionary
    """

    user = None
    response = self.user_show(uid, allattrs=True)

    if response.json()['result']:
        user = delist(response.json()['result']['result'])

    return user


def users(
        self,
        searchstring: Union[str, None]=None,
        uid: Union[str, None]=None,
        uidnumber: Union[int, None]=None,
        in_group: Union[str, list, None]=None,
        mail: Union[str, list, None]=None,
):
    """
    @brief      Given a some search parameters find all the user acounts that match

    @param      self          The object
    @param      searchstring  The searchstring used on any otptoken attribute
    @param      uniqueid      substring used to match otptoken uniqueid
    @param      owner         search for tokens owned but specified user

    @return     { description_of_the_return_value }
    """

    response = self.user_find(
        searchstring=searchstring,
        uid=uid,
        uidnumber=uidnumber,
        in_group=in_group,
        mail=mail,
        allattrs=True
    )

    userlist = []

    if response.json()['result']:
        userlist = listdelist(response.json()['result']['result'])

    return userlist


def userlist(
        self,
        uids: Union[str, list, None]=None,
        groups: Union[str, list, None]=None
):
    """
    @brief      Given a list of uids and/or groups, return a list of usernames that match or are members

    @param      self          The object
    @param      searchstring  The searchstring used on any otptoken attribute
    @param      uniqueid      substring used to match otptoken uniqueid
    @param      owner         search for tokens owned but specified user

    @return     { description_of_the_return_value }
    """

    userlist = []

    if uids or groups:
        if uids:
            if isinstance(uids, list):
                for username in uids:
                    response = self.users(uid=username)
                    for user in response:
                        userlist.append(user['uid'])
            else:
                response = self.users(uid=uids)
                for user in response:
                    userlist.append(user['uid'])
        if groups:
            if isinstance(groups, list):
                for groupname in groups:
                    response = self.users(in_group=groupname)
                    for user in response:
                        userlist.append(user['uid'])
            else:
                response = self.users(in_group=groups)
                for user in response:
                    userlist.append(user['uid'])
    else:
        for user in self.users():
            userlist.append(user['uid'])

    return sorted(set(userlist))

def user_show(
        self,
        uid: str,
        rights: Union[bool, None]=None,
        allattrs: Union[bool, None]=None,
        raw: Union[bool, None]=None
):
    """
    @brief      A complete implementation of the user_show command

    @param      self  The object

    @param      uid of the user to be shown

    @param      rights, if true, displays the access rights of this user

    @param      all, retrieves all attributes

    @param      raw, returns the raw response, only changes output format

    @return     the requests.Response from the ping request
    """

    method = 'user_show'

    args = uid

    params = {}

    if rights is not None:
        params['rights'] = rights

    if allattrs is not None:
        params['all'] = allattrs

    if raw is not None:
        params['raw'] = raw

    return self.request(
        method,
        args=args,
        params=params
    )


def user_find(
        self,
        searchstring: Union[str, None]=None,
        uid: Union[str, None]=None,
        uidnumber: Union[int, None]=None,
        in_group: Union[str, list, None]=None,
        mail: Union[str, list, None]=None,
        rights: Union[bool, None]=None,
        allattrs: Union[bool, None]=None,
        raw: Union[bool, None]=None
):
    """
    @brief      A partial implementation of the user_find request

    @param      self  The object

    @param      uid of the user to be shown

    @param      rights, if true, displays the access rights of this user

    @param      all, retrieves all attributes

    @param      raw, returns the raw response, only changes output format

    @return     the requests.Response from the ping request
    """

    method = 'user_find'

    args = None

    if searchstring:
        args = searchstring

    params = {}

    if uid is not None:
        params['uid'] = uid

    if uidnumber is not None:
        params['uidnumber'] = uidnumber

    if in_group is not None:
        params['in_group'] = in_group

    if mail is not None:
        params['mail'] = mail

    if rights is not None:
        params['rights'] = rights

    if allattrs is not None:
        params['all'] = allattrs

    if raw is not None:
        params['raw'] = raw

    return self.request(
        method,
        args=args,
        params=params
    )


def user_getattr(
        self,
        uid: type=str,
        attribute: type=str
):
    user = self.user(uid=uid)

    attributevalue = None

    if attribute in user:
        attributevalue = user[attribute]

    return attributevalue


# Write methods that may cause changes to the directory
# These methods MUST require dryrun to be false to write changes
def user_mod(
        self,
        # Arguments
        uid: type=str,
        # Options
        givenname: Union[str, None]=None,
        sn: Union[str, None]=None,
        cn: Union[str, None]=None,
        displayname: Union[str, None]=None,
        initials: Union[str, None]=None,
        homedirectory: Union[str, None]=None,
        gecos: Union[str, None]=None,
        loginshell: Union[str, None]=None,
        krbprincipalname: Union[str, None]=None,
        krbprincipalexpiration: Union[datetime, None]=None,
        krbpasswordexpiration: Union[datetime, None]=None,
        mail: Union[str, None]=None,
        userpassword: Union[str, None]=None,
        random: Union[bool, None]=None,
        uidnumber: Union[int, None]=None,
        gidnumber: Union[int, None]=None,
        telephonenumber: Union[str, None]=None,
        mobile: Union[str, None]=None,
        ou: Union[str, None]=None,
        title: Union[str, None]=None,
        manager: Union[str, None]=None,
        carlicense: Union[str, None]=None,
        ipasshpubkey: Union[str, None]=None,
        ipauserauthtype: Union[str, None]=None,
        userclass: Union[str, None]=None,
        ipatokenradiusconfiglink: Union[str, None]=None,
        ipatokenradiususername: Union[str, None]=None,
        departmentnumber: Union[str, None]=None,
        employeenumber: Union[str, None]=None,
        employeetype: Union[str, None]=None,
        preferredlanguage: Union[str, None]=None,
        nsaccountlock: Union[str, None]=None,
        no_members: Union[bool, None]=None,
        rename: Union[str, None]=None,
        allattrs: Union[bool, None]=None,
        raw: Union[bool, None]=None,
        addattr: Union[dict, None]=None,
        setattr: Union[dict, None]=None,
        delattr: Union[dict, None]=None
):

    method = 'user_mod'

    args = uid

    caseparams = {}

    params = {}

    currentuser = self.user(uid)

    # If there is no current user then these tests will break,
    # and the proposed actions are pointless so...
    if currentuser:

        # Case insensitive!
        if givenname is not None:
            params['givenname'] = givenname
            if givenname.lower() == currentuser['givenname'].lower():
                caseparams['givenname'] = "case_%s" % currentuser['givenname']

        if sn is not None:
            params['sn'] = sn
            if sn.lower() == currentuser['sn'].lower():
                caseparams['sn'] = "case_%s" % currentuser['sn']

        # Case insensitive!
        if cn is not None:
            params['cn'] = cn
            if cn.lower() == currentuser['cn'].lower():
                caseparams['cn'] = "case_%s" % currentuser['cn']

        # Case insensitive!
        if displayname is not None:
            params['displayname'] = displayname
            if displayname.lower() == currentuser['displayname'].lower():
                caseparams['displayname'] = "case_%s" % currentuser['displayname']

        # Case insensitive! Should be .upper
        if initials is not None:
            params['initials'] = initials
            if initials.lower() == currentuser['initials'].lower():
                caseparams['initials'] = "case_%s" % currentuser['initials']

        # Case insensitive!
        if homedirectory is not None:
            params['homedirectory'] = homedirectory
            if homedirectory.lower() == currentuser['homedirectory'].lower():
                caseparams['homedirectory'] = "case_%s" % currentuser['homedirectory']

        # Case insensitive!
        if gecos is not None:
            params['gecos'] = gecos
            if gecos.lower() == currentuser['gecos'].lower():
                caseparams['gecos'] = "case_%s" % currentuser['gecos']

        if loginshell is not None:
            params['loginshell'] = loginshell

        if krbprincipalname is not None:
            params['krbprincipalname'] = krbprincipalname

        if krbprincipalexpiration is not None:
            params['krbprincipalexpiration'] = krbprincipalexpiration.strftime("%Y%m%d%H%M%SZ")

        if krbpasswordexpiration is not None:
            params['krbpasswordexpiration'] = krbpasswordexpiration.strftime("%Y%m%d%H%M%SZ")

        # Case insensitive!
        if mail is not None:
            params['mail'] = mail
            if mail.lower() == currentuser['mail'].lower():
                caseparams['mail'] = "case_%s" % currentuser['mail']

        if userpassword is not None:
            params['userpassword'] = userpassword

        if random is not None:
            params['random'] = random

        if uidnumber is not None:
            params['uidnumber'] = uidnumber

        if gidnumber is not None:
            params['gidnumber'] = gidnumber

        if telephonenumber is not None:
            params['telephonenumber'] = telephonenumber

        if mobile is not None:
            params['mobile'] = mobile

        if ou is not None:
            params['ou'] = ou

        # Case insensitive!
        if title is not None:
            params['title'] = title
            if title.lower() == currentuser['title'].lower():
                caseparams['title'] = "case_%s" % currentuser['title']

        # Case insensitive!
        if manager is not None:
            params['manager'] = manager
            if manager.lower() == currentuser['manager'].lower():
                caseparams['manager'] = "case_%s" % currentuser['manager']

        if carlicense is not None:
            params['carlicense'] = carlicense

        if ipasshpubkey is not None:
            params['ipasshpubkey'] = ipasshpubkey

        if ipauserauthtype is not None:
            params['ipauserauthtype'] = ipauserauthtype

        if userclass is not None:
            params['userclass'] = userclass

        if ipatokenradiusconfiglink is not None:
            params['ipatokenradiusconfiglink'] = ipatokenradiusconfiglink

        if ipatokenradiususername is not None:
            params['ipatokenradiususername'] = ipatokenradiususername

        # Case insensitive!
        if departmentnumber is not None:
            params['departmentnumber'] = departmentnumber
            if departmentnumber.lower() == currentuser['departmentnumber'].lower():
                caseparams['departmentnumber'] = "case_%s" % currentuser['departmentnumber']

        if employeenumber is not None:
            params['employeenumber'] = employeenumber

        # Case insensitive!
        if employeetype is not None:
            params['employeetype'] = employeetype
            if employeetype.lower() == currentuser['employeetype'].lower():
                caseparams['employeetype'] = "case_%s" % currentuser['employeetype']

        # Case insensitive!
        if preferredlanguage is not None:
            params['preferredlanguage'] = preferredlanguage
            if preferredlanguage.lower() == currentuser['prefeif preferredlanguage'].lower():
                caseparams['prefeif preferredlanguage'] = "case_%s" % currentuser['prefeif preferredlanguage']

        if nsaccountlock is not None:
            params['nsaccountlock'] = nsaccountlock

        if no_members is not None:
            params['no_members'] = no_members

        if rename is not None:
            params['rename'] = rename

        if allattrs is not None:
            params['all'] = allattrs

        if raw is not None:
            params['raw'] = raw

        # Use addattr, setattr and delatter to modify custom variables
        if setattr is not None:
            for key, value in setattr.items():
                if params.get('setattr'):
                    params['setattr'].append(
                        "%s=%s" % (
                            key,
                            value
                        )
                    )
                else:
                    params['setattr']=[
                        "%s=%s" % (
                            key,
                            value
                        )
                    ]
                # Check for case insensitivity
                if currentuser.get(key):
                    if currentuser[key].lower() == value.lower():
                        if caseparams.get('setattr'):
                            caseparams['setattr'][key] = value
                        else:
                            caseparams['setattr']={
                                key: value
                            }

        if addattr is not None:
            for key, value in addattr.items():
                if params.get('addattr'):
                    params['addattr'].append(
                        "%s=%s" % (
                            key,
                            value
                        )
                    )
                else:
                    params['addattr']=[
                        "%s=%s" % (
                            key,
                            value
                        )
                    ]

        if delattr is not None:
            for key, value in delattr.items():
                if params.get('delattr'):
                    params['delattr'].append(
                        "%s=%s" % (
                            key,
                            value
                        )
                    )
                else:
                    params['delattr']=[
                        "%s=%s" % (
                            key,
                            value
                        )
                    ]

    # If theres a case issue detected, do another user_mod to
    # set the case insenistive values to a dummy value
    # this might fail if the strings end up too long
    if caseparams:
        self.user_mod(
            uid,
            # Need to default to None if not defined!
            givenname=caseparams.get('givenname'),
            cn=caseparams.get('cn'),
            displayname=caseparams.get('displayname'),
            initials=caseparams.get('initials'),
            homedirectory=caseparams.get('homedirectory'),
            gecos=caseparams.get('gecos'),
            mail=caseparams.get('mail'),
            title=caseparams.get('title'),
            manager=caseparams.get('manager'),
            departmentnumber=caseparams.get('departmentnumber'),
            employeetype=caseparams.get('employeetype'),
            preferredlanguage=caseparams.get('preferredlanguage')
        )

    # This is a write method so, prepare the request
    prepared = self.preprequest(
        method,
        args=args,
        params=params
    )

    # then check if it's a dryrun before executing it
    if not self._dryrun:
        response = self._session.send(prepared)
    else:
        response = prepared

    return response



"""
Methods for otptokens
otp - one time password, used for two factor & multifactor authentication.
"""
from typing import Union
from datetime import datetime


def otptoken_find(
        self,
        searchstring: Union[str, None]=None,
        uniqueid: Union[str, None]=None,
        owner: Union[str, None]=None,
        no_members: Union[bool, None]=None,
        allattrs: Union[bool, None]=None,
        raw: Union[bool, None]=None
):
    """
    @brief Partial implementation the otptoken_show request, only searches by parameters below

    @param self The object
    @param uniqueid, the unique identifier (`iparokenuniqueid`) attribute of the token to find
    @param owner, the owner (`iparokenowner`) attribute of the token to find
    @param no_members, if true this suppresses processing of membershib attributes
    @param all, retrieves all attributes
    @param raw, returns the raw response, only changes output format

    @return the request.Response from the otpoken_show request
    """

    method = 'otptoken_find'

    args = None

    if searchstring:
        args = searchstring

    params = {}

    if uniqueid is not None:
        params['ipatokenuniqueid'] = uniqueid

    if owner is not None:
        params['ipatokenowner'] = owner

    if no_members is not None:
        params['no_members'] = no_members

    if allattrs is not None:
        params['all'] = allattrs

    if raw is not None:
        params['raw'] = raw

    return self.request(
        method,
        args=args,
        params=params
    )


def otptoken_show(
        self,
        uniqueid: str,
        no_members: Union[bool, None]=None,
        allattrs: Union[bool, None]=None,
        raw: Union[bool, None]=None
):
    """
    @brief Complete implementation the otptoken_show request

    @param self The object
    @param uniqueid, the unique identifier (`iparokenuniqueid`) attribute of the token to show
    @param no_members, if true this suppresses processing of membershib attributes
    @param all, retrieves all attributes
    @param raw, returns the raw response, only changes output format

    @return the request.Response from the otpoken_show request
    """

    method = 'otptoken_show'

    args = uniqueid

    params = {}

    if no_members is not None:
        params['no_members'] = no_members

    if allattrs is not None:
        params['all'] = allattrs

    if raw is not None:
        params['raw'] = raw

    return self.request(
        method,
        args=args,
        params=params
    )


# Write methods that may cause changes to the directory
# These methods MUST require dryrun to be false to write changes
def otptoken_add(
        self,
        # Arguments
        uniqueid: str,
        # Options
        otptype: Union[str, None]=None,
        description: Union[str, None]=None,
        owner: Union[str, None]=None,
        disabled: Union[bool, None]=None,
        notbefore: Union[datetime, None]=None,
        notafter: Union[datetime, None]=None,
        vendor: Union[str, None]=None,
        model: Union[str, None]=None,
        serial: Union[str, None]=None,
        otpkey: Union[str, None]=None,
        otpalgorithm: Union[str, None]=None,
        otpdigits: Union[int, None]=None,
        otpclockoffset: Union[int, None]=None,
        otptimestep: Union[int, None]=None,
        otpcounter: Union[int, None]=None,
        no_qrcode: Union[bool, None]=None,
        no_members: Union[bool, None]=None,
        otpsetattr: Union[list, None]=None,
        addattr: Union[list, None]=None,
        # Custom
        managedby: Union[list, None]=None
):

    method = 'otptoken_add'

    args = uniqueid

    params = {}

    # These options need some checking before submitting a request
    typevalues = ['totp', 'hotp', 'TOTP', 'HOTP']
    if otptype is not None:
        if otptype in typevalues:
            params['type'] = otptype
        else:
            raise ValueError(
                "otptoken_add: otptype must be one of %r" % typevalues
            )

    algovalues = ['sha1', 'sha256', 'sha384', 'sha512']
    if otpalgorithm is not None:
        if otpalgorithm.lower() in algovalues:
            params['ipatokenotpalgorithm'] = otpalgorithm.lower()
        else:
            raise ValueError(
                "otptoken_add: otpalgorithm must be one of %r" % typevalues
            )

    # These options just need to be mapped to request parameters
    if description is not None:
        params['description'] = description

    if owner is not None:
        params['ipatokenowner'] = owner

    if disabled is not None:
        params['ipatokendisabled'] = disabled

    if notafter is not None:
        params['ipatokennotafter'] = notafter

    if notbefore is not None:
        params['ipatokennotbefore'] = notbefore

    if vendor is not None:
        params['ipatokenvendor'] = vendor

    if model is not None:
        params['ipatokenmodel'] = model

    if serial is not None:
        params['ipatokenserial'] = serial

    if otpkey is not None:
        params['ipatokenotpkey'] = otpkey

    if otpdigits is not None:
        params['ipatokenotpdigits'] = otpdigits

    if otpclockoffset is not None:
        params['ipatokentotpclockoffset'] = otpclockoffset

    if otptimestep is not None:
        params['ipatokentotptimestep'] = otptimestep

    if otpcounter is not None:
        params['ipatokenhotpcounter'] = otpcounter

    if no_qrcode is not None:
        params['no_qrcode'] = no_qrcode

    if no_members is not None:
        params['no_members'] = no_members

    if otpsetattr is not None:
        params['setattr'] = otpsetattr

    if addattr is not None:
        params['addattr'] = addattr
    # Custom
    if managedby is not None:
        print("I'm special")

    prepared = self.preprequest(
        method,
        args=args,
        params=params
    )

    if not self._dryrun:
        response = self._session.send(prepared)
    else:
        response = prepared

    return response


def otptoken_add_managedby(
        self,
        # Arguments
        uniqueid: str,
        user: Union[list, str],
        no_members: Union[bool, None]=None,
        allattrs: Union[bool, None]=None,
        raw: Union[bool, None]=None
):
    """
    @brief adds a user to the managedBy attributes

    @return the response from the otptoken_add_managedby request, unless dry run where it returns the prepared response
    """

    method = 'otptoken_add_managedby'

    args = uniqueid

    if not isinstance(user, list):
        user = [user]

    params = {
        'user': user
    }

    if no_members is not None:
        params['no_members'] = no_members

    if allattrs is not None:
        params['all'] = allattrs

    if raw is not None:
        params['raw'] = raw

    prepared = self.preprequest(
        method,
        args=args,
        params=params
    )

    if not self._dryrun:
        response = self._session.send(prepared)
    else:
        response = prepared

    return response


def otptoken_remove_managedby(
        self,
        # Arguments
        uniqueid: str,
        user: Union[list, str],
        no_members: Union[bool, None]=None,
        allattrs: Union[bool, None]=None,
        raw: Union[bool, None]=None
):
    """
    @brief removes a user to the managedBy attributes

    @return the response from the otptoken_add_managedby request, unless dry run where it returns the prepared response
    """

    method = 'otptoken_remove_managedby'

    args = uniqueid

    if not isinstance(user, list):
        user = [user]

    params = {
        'user': user
    }

    if no_members is not None:
        params['no_members'] = no_members

    if allattrs is not None:
        params['all'] = allattrs

    if raw is not None:
        params['raw'] = raw

    prepared = self.preprequest(
        method,
        args=args,
        params=params
    )

    if not self._dryrun:
        response = self._session.send(prepared)
    else:
        response = prepared

    return response

# The next methods produce objects, or lists or objects
# These are intended for simple operations that don't require
# full feedback from the response objects
# e.g. won't say why a request found nothing


def otptoken(
        self,
        uniqueid: str
):
    """
    @brief      { function_description }

    @param      self      The object
    @param      uniqueid  The uniqueid of he token to be returned

    @return     { description_of_the_return_value }
    """

    response = self.otptoken_show(uniqueid, allattrs=True)

    if response.json()['result']:
        return response.json()['result']['result']
    else:
        return None


def otptokens(
        self,
        searchstring: Union[str, None]=None,
        uniqueid: Union[str, None]=None,
        owner: Union[str, None]=None
):
    """
    @brief      { function_description }

    @param      self          The object
    @param      searchstring  The searchstring used on any otptoken attribute
    @param      uniqueid      substring used to match otptoken uniqueid
    @param      owner         search for tokens owned but specified user

    @return     { description_of_the_return_value }
    """

    response = self.otptoken_find(
        searchstring=searchstring,
        uniqueid=uniqueid,
        owner=owner,
        allattrs=True
    )

    if response.json()['result']:
        return response.json()['result']['result']
    else:
        return []



"""
{ item_description }
"""
import json
from typing import Union
from datetime import datetime
import requests
import urllib3
from ._utils import jsonprepreq


class Api:
    """
    @brief      Class for api connection to an IPA server.
    """
    # Import local methods
    from ._user import (
        user_show,
        user_find,
        user,
        users,
        userlist,
        user_getattr,
        user_mod
    )
    from ._group import (
        group_show,
        group_find,
        group,
        groups,
        grouplist
    )
    from ._otptoken import (
        otptoken_find,
        otptoken_show,
        otptoken,
        otptokens,
        otptoken_remove_managedby,
        otptoken_add_managedby,
        otptoken_add
    )

    def __init__(
            self,
            host: type=str,
            username: type=str,
            password: type=str,
            port: int=443,
            protocol: str='https',
            verify_ssl: bool=True,
            verify_method: Union[bool, str]=True,
            verify_warnings: bool=True,
            version: str='2.228',
            dryrun: bool=False
    ):
        """
        @brief The initiator of the pyfreeipa.Api class

        @param self This object
        @param host The address of the IPA server
        @param username The username of the IPA user to connect to the IPA server
        @param password The password for the IPA user
        @param port The port to connect to (Default: 443)
        @param protocol The protocol to connect with (Default: https)
        @param verify_ssl If True then the TLS/SSL connection will verified as being secure and has valid certificates. (Default: True)
        @param verify_method Can be a True, False, or a string pointing to a certificate or directory holding certificaates. (Default: True)
        @param verify_warnings If True then TLS/SSL warnings will be emitted to stderr. (Default: True)
        @param version Used to over-ride the default IPA API version string. (Default: 2.228)
        @param dryrun If set to True this will stop any requests from making changes to the IPA directory (Default: False)
        """

        self._host = host
        self._username = username
        self._password = password
        self._port = port
        self._protocol = protocol
        self._verify_ssl = verify_ssl
        self._verify_method = verify_method
        self._verify_warnings = verify_warnings
        self._version = version
        self._dryrun = dryrun

        self.warnings = []

        if not self._verify_warnings:
            reason = (
                'Verifying TLS connection to %s disabled.' %
                self._host
            )
            self.warnings.append(reason)

        self._baseurl = (
            "%s://%s/ipa" %
            (
                self._protocol,
                self._host
            )
        )

        self._sessionurl = "%s/session/json" % self._baseurl
        self._loginurl = "%s/session/login_password" % self._baseurl

        if not self._verify_warnings:
            reason = (
                "TLS warnings from %s disabled" %
                self._host
            )
            self.warnings.append(reason)
            urllib3.disable_warnings()

        self._session = requests.Session()
        self._session.url = self._baseurl
        self._session.verify = self._verify_ssl
        self._session.headers = {
            'Referer': self._baseurl,
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }

        self.login()

    def get(
            self,
            *dargs,
            **kwargs
    ):
        """
        @brief This exposes a raw get method for the session
        """
        response = self._session.get(
            *dargs,
            **kwargs
        )
        return response

    def request(
            self,
            method: str,
            args=None,
            params=None
    ):
        """
        @brief This cleans up the args and parameters posts the and handles
        the request and returns the response

        @param self This object
        @param method The FreeIPA API method to request
        @param args A list of arguments to be passed to the method request
        @param params A dictionary of parameters to be passed to the method request

        @return a requests.Response object
        """
        if args:
            if not isinstance(args, list):
                args = [args]
        else:
            args = []

        if not params:
            params = {}

        if self._version:
            params.setdefault('version', self._version)

        data = {
            'id': 0,
            'method': method,
            'params': [args, params]
        }

        response = self.post(
            self._sessionurl,
            data=json.dumps(data, default=str)
        )

        return response

    def preprequest(
            self,
            method: str,
            args=None,
            params=None
    ):
        """
        @brief This cleans up the args and parameters posts the and creates a prepared reques from the internal sessions object

        @param self This object
        @param method The FreeIPA API method to request
        @param args A list of arguments to be passed to the method request
        @param params A dictionary of parameters to be passed to the method request

        @return a requests.Response object
        """
        if args:
            if not isinstance(args, list):
                args = [args]
        else:
            args = []

        if not params:
            params = {}

        if self._version:
            params.setdefault('version', self._version)

        data = {
            'id': 0,
            'method': method,
            'params': [args, params]
        }

        request = requests.Request(
            'POST',
            self._sessionurl,
            data=json.dumps(data, default=str)
        )

        return self._session.prepare_request(request)

    def clearwarnings(self):
        """
        @brief      Retrieve and clear the warning array

        @param      self  The object

        @return     the warnings array before it was cleared
        """
        warnings = self.warnings
        self.warnings = []
        return warnings

# All definitions from this point are IPA API commands
# Two sections Read and Write, with methods in alphabetical order

# Read methods that make no change to the directory and should work in

    def login(self):
        """
        @brief      Returns the response from the login command

        @param      self  The object

        @return     the requests.Response from the login request
        """
        logindata = {
            'user': self._username,
            'password': self._password
        }

        loginheaders = {
            'referer': self._loginurl,
            'Content-Type': 'application/x-www-form-urlencoded',
            'Accept': 'application/json'
        }

        response = self.post(
            self._loginurl,
            data=logindata,
            headers=loginheaders
        )

        return response


    def ping(self):
        """
        @brief      Returns the response from the ping command

        @param      self  The object

        @return     the requests.Response from the ping request
        """
        return self.request('ping')


    def whoami(self):
        """
        @brief      Returns the response from the whoami command

        @param      self  The object

        @return     the requests.Response from the whoami request
        """
        return self.request('whoami')

## WRITE METHODS: Methods beyond this point can update spectrumscale
## these methods MUST make no changes if self._dryrun is True
##

## Sends a prepared request
    def send(
        self,
        preprequest: type=requests.PreparedRequest
    ):
        response = None
        if self._dryrun:
            response = jsonprepreq(preprequest)
            response['dryrun'] = True
        else:
            response = self._session.send(preprequest)

        return response

## POST a request directly
    def post(
            self,
            *dargs,
            **kwargs
    ):
        """
        @brief This exposes a raw post method for the internal session
        """

        response = self._session.post(
            *dargs,
            **kwargs
        )

        return response

# PUT a request directly
    def put(
            self,
            *dargs,
            **kwargs
    ):
        """
        @brief This exposes a raw put method for the internal session
        """

        response = self._session.put(
            *dargs,
            **kwargs
        )

        return response



"""
Some common utilities for the API
"""
import requests
import json
from datetime import datetime
from typing import List



def delist(
    thedict: dict
):
    """
    @brief      Given a dictionary, check all the key values and if
                the value is a list, and the list has one item, then
                replace the list with the single item in the list

    @param      thedict  The dictionary to process

    @return     a dictionary with no single element lists
    """
    cleandict = {}
    timeformat = '%Y%m%d%H%M%SZ'
    for key in thedict:
        cleandict[key] = thedict[key]
        if isinstance(thedict[key], list):
            if len(thedict[key]) == 1:
                cleandict[key] = thedict[key][0]

        # Convert serialised datetime dicts back into a datetime object
        if isinstance(cleandict[key], dict):
            if '__datetime__' in cleandict[key]:
                cleandict[key] = datetime.strptime(
                    cleandict[key]['__datetime__'],
                    timeformat
                )

    return cleandict


def listdelist(
    thelist: List[dict]
):
    """
    @brief      Uses the delist functio on a list of Dictioaries

    @param      thelist  The list of dictionaries

    @return     a list of dictionaries with no single element lists
    """
    cleanlist = []
    for item in thelist:
        cleanlist.append(delist(item))

    return cleanlist

def jsonprepreq(
    preprequest: type=requests.PreparedRequest
):
    """
    At this point it is completely built and ready
    to be fired; it is "prepared". but we need it as JSONable dict
    """
    jsonprepreq = {
        'headers': {},
        'body': {}
    }

    jsonprepreq['method'] = preprequest.method
    jsonprepreq['url'] = preprequest.url
    for key, value in preprequest.headers.items():
        jsonprepreq['headers'][key] = value

    jsonprepreq['body'] = json.loads(preprequest.body)
    return jsonprepreq



"""
A generic wrapper script for the pyfreeipa Api class
"""
import json
import sys
from pyfreeipa.Api import Api
from pyfreeipa.configuration import CONFIG


def main():
    """
    @brief      This provides a wrapper for the pyfreeipa module

    @return     { description_of_the_return_value }
    """

    if CONFIG['command'] == 'dumpconfig':
        print(json.dumps(CONFIG, indent=2, sort_keys=True))
        sys.exit(0)

    # Define API session
    ipaapi = Api(
        host=CONFIG['ipaserver']['host'],
        username=CONFIG['ipaserver']['user'],
        password=CONFIG['ipaserver']['password'],
        port=CONFIG['ipaserver']['port'],
        verify_ssl=CONFIG['ipaserver']['verify_ssl'],
        verify_method=CONFIG['ipaserver']['verify_method'],
        verify_warnings=CONFIG['ipaserver']['verify_warnings'],
        dryrun=CONFIG['dryrun']
    )

    if CONFIG['command'] == 'connectiontest':
        print(
            "Test connection to %s" %
            CONFIG['ipaserver']['host']
        )

        response = ipaapi.ping()
        if response.ok:
            print(
                'Successfully pinged as %s on %s' %
                (
                    CONFIG['ipaserver']['user'],
                    CONFIG['ipaserver']['host']
                )
            )
            result = response.json()['result']
            print(result['summary'])
        else:
            print(
                'Failed to ping as %s in to %s, reason "%s: %s"' %
                (
                    CONFIG['ipaserver']['user'],
                    CONFIG['ipaserver']['host'],
                    response.status_code,
                    response.reason
                )
            )

    else:
        print("No command provided")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
Install module
"""

import setuptools

setuptools.setup(
    name="pyfreeipa",
    description="Python FreeIPA API interface",
    author="Aaron Hicks",
    author_email="aethylred@gmail.com",
    version="0.2.0",
    packages=setuptools.find_packages(),
    license='Apache 2.0',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'certifi',
        'chardet',
        'idna',
        'PyYAML',
        'requests',
        'typing',
        'urllib3',
    ]
)



"""
Process command line arguments and/or load configuration file
mostly used by the test scripts
"""
import argparse
import sys
import os.path
from typing import Union
import yaml


def do_args():
    """
    @brief      { function_description }

    @return     { description_of_the_return_value }
    """
    # Parse command line arguments and modify config
    parser = argparse.ArgumentParser(
        prog='pyfreeipa.py',
        description='Python FreeIPA tools'
    )

    # Command line arguments
    parser.add_argument(
        "-v",
        "--verbose",
        dest='verbose',
        help="Increase output to stderr and stdout",
        action="store_true"
    )

    parser.add_argument(
        "-q",
        "--quiet",
        dest='quiet',
        help="Reduce output to stderr and stdout",
        action="store_true"
    )

    parser.add_argument(
        "-d",
        "--dry_run",
        dest='dryrun',
        help="Do a dry run, no changes written to IPA server",
        action="store_true"
    )

    parser.add_argument(
        "-f",
        "--file",
        default='pyfreeipa.conf.yaml',
        dest='file',
        help="Specify a configuration file",
    )

    parser.add_argument(
        '-s',
        '--server',
        default=None,
        type=str,
        dest='server',
        help="Hostname of IPA server"
    )

    parser.add_argument(
        '-u',
        '--user',
        default=None,
        type=str,
        dest='user',
        help="The username used to connect to the IPA server"
    )

    parser.add_argument(
        '-p',
        '--password',
        default=None,
        type=str,
        dest='password',
        help="The password used to conenct to the IPA server"
    )

    parser.add_argument(
        '--uid',
        default=None,
        type=str,
        nargs='+',
        dest='uid',
        help=(
            "Passes a unique identifier, such as a uid of a user,"
            "or the uniqueid of an otptoken. Multiple uids can be specified."
        )
    )

    parser.add_argument(
        '--group',
        default=None,
        type=str,
        nargs='+',
        dest='group',
        help="The name of a group to use, multiple groups can be specified"
    )

    parser.add_argument(
        '--port',
        default=None,
        type=str,
        dest='port',
        help="The password used to conenct to the IPA server"
    )

    parser.add_argument(
        '--version',
        default=None,
        type=str,
        dest='version',
        help="The IPA server API version"
    )

    parser.add_argument(
        '--verify_ssl',
        default=None,
        type=bool,
        dest='verify_ssl',
        help=(
            "If true the SSL certificate of the"
            " IPA server will be verified"
        )
    )

    parser.add_argument(
        '--verify_warnings',
        default=None,
        type=bool,
        dest='verify_warnings',
        help=(
            "If false warnings about the SSL state of "
            "the IPA server will be silenced"
        )
    )

    parser.add_argument(
        '--verify_method',
        default=None,
        type=Union[bool, str],
        dest='verify_method',
        help=(
            "The method used to validate the SSL state of "
            "the IPA server"
        )
    )

    # Positional commands
    parser.add_argument(
        dest='command',
        help='Command help',
        default=None,
        nargs='?',
        type=str,
        choices=[
            'dumpconfig',
            'connectiontest'
        ]
    )

    return parser.parse_args()

# Create the CONFIG to be imported elsewhere
# Set defaults
CONFIG = {
    'ipaserver': {
        'host': 'ipaserver.example.org',
        'user': 'username',
        'password': None,
        'port': 443,
        'version': 2.228,
        'verify_ssl': True,
        'verify_method': True,
        'verify_warnings': True
    },
    'otptoken': {
        'managedby': [],
        'ownermanagedby': True
    }
}


ARGS = do_args()

# Override configuration defaults with values from the config file
if os.path.isfile(ARGS.file):
    with open(ARGS.file, 'r') as configfile:
        CONFIG.update(yaml.load(configfile))

# Override configuration loaded from file with command line arguments
if ARGS.server:
    CONFIG['ipaserver']['host'] = ARGS.server

if ARGS.user:
    CONFIG['ipaserver']['user'] = ARGS.user

if ARGS.password:
    CONFIG['ipaserver']['password'] = ARGS.password

if ARGS.port:
    CONFIG['ipaserver']['port'] = ARGS.port

if ARGS.version:
    CONFIG['ipaserver']['version'] = ARGS.version

# This one can be bool or str values
if ARGS.verify_method is not None:
    CONFIG['ipaserver']['verify_method'] = ARGS.verify_method

if ARGS.verify_ssl is not None:
    CONFIG['ipaserver']['verify_ssl'] = ARGS.verify_ssl

if ARGS.verify_warnings is not None:
    CONFIG['ipaserver']['verify_warnings'] = ARGS.verify_warnings

# If there's no config file, write one
if not os.path.isfile(ARGS.file):
    print(
        "The configuration file %s was missing,"
        " wrote default configuration to file" %
        ARGS.file
    )
    with open(ARGS.file, 'w') as configfile:
        yaml.dump(vars(CONFIG), configfile, default_flow_style=False)
    sys.exit(0)

# Set state from command line
CONFIG['command'] = ARGS.command
CONFIG['dryrun'] = ARGS.dryrun
CONFIG['groups'] = None
if ARGS.group:
    cleangroups = []
    for group in ARGS.group:
        if ',' in group:
            for splitgroup in group.split(','):
                cleangroups.append(splitgroup)
        else:
            cleangroups.append(group)

    CONFIG['groups'] = sorted(set(cleangroups))

CONFIG['users'] = None
if ARGS.uid:
    cleanusers = []
    for user in ARGS.uid:
        if ',' in user:
            for splituser in user.split(','):
                cleanusers.append(splituser)
        else:
            cleanusers.append(user)

    CONFIG['users'] = sorted(set(cleanusers))

if CONFIG['otptoken']['managedby']:
    if not isinstance(CONFIG['otptoken']['managedby'], list):
        CONFIG['otptoken']['managedby'] = [CONFIG['otptoken']['managedby']]
